// Microsoft Visual C++ generated resource script.
//
#include "resource.h"

#define APSTUDIO_READONLY_SYMBOLS
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 2 resource.
//
#include <winresrc.h>

/////////////////////////////////////////////////////////////////////////////
#undef APSTUDIO_READONLY_SYMBOLS

/////////////////////////////////////////////////////////////////////////////
// English (U.S.) resources

#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_ENU)
#ifdef _WIN32
LANGUAGE LANG_ENGLISH, SUBLANG_ENGLISH_US
#pragma code_page(1252)
#endif //_WIN32

#ifdef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// TEXTINCLUDE
//

1 TEXTINCLUDE 
BEGIN
    "resource.h\0"
END

2 TEXTINCLUDE 
BEGIN
    "#include <winresrc.h>\r\n"
    "\0"
END

3 TEXTINCLUDE 
BEGIN
    "\r\n"
    "\0"
END

#endif    // APSTUDIO_INVOKED


/////////////////////////////////////////////////////////////////////////////
//
// Icon
//

// Icon with lowest ID value placed first to ensure application icon
// remains consistent on all systems.
IDI_DESKFLOW             ICON                    "deskflows.ico"
IDI_TASKBAR_NOT_RUNNING ICON                    "tb_idle.ico"
IDI_TASKBAR_NOT_WORKING ICON                    "tb_error.ico"
IDI_TASKBAR_NOT_CONNECTED ICON                    "tb_wait.ico"
IDI_TASKBAR_CONNECTED   ICON                    "tb_run.ico"

/////////////////////////////////////////////////////////////////////////////
//
// Menu
//

IDR_TASKBAR MENU 
BEGIN
    POPUP "Deskflow"
    BEGIN
        MENUITEM "Show Status",                 IDC_TASKBAR_STATUS
        MENUITEM "Show Log",                    IDC_TASKBAR_SHOW_LOG
        MENUITEM "Copy Log To Clipboard",       IDC_TASKBAR_LOG
        POPUP "Set Log Level"
        BEGIN
            MENUITEM "Error",                       IDC_TASKBAR_LOG_LEVEL_ERROR
            MENUITEM "Warning",                     IDC_TASKBAR_LOG_LEVEL_WARNING
            MENUITEM "Note",                        IDC_TASKBAR_LOG_LEVEL_NOTE
            MENUITEM "Info",                        IDC_TASKBAR_LOG_LEVEL_INFO
            MENUITEM "Debug",                       IDC_TASKBAR_LOG_LEVEL_DEBUG
            MENUITEM "Debug1",                      IDC_TASKBAR_LOG_LEVEL_DEBUG1
            MENUITEM "Debug2",                      IDC_TASKBAR_LOG_LEVEL_DEBUG2
        END
        MENUITEM "Reload Configuration",        IDC_RELOAD_CONFIG
        MENUITEM "Force Reconnect",             IDC_FORCE_RECONNECT
        MENUITEM "Reset Server",                ID_DESKFLOW_RESETSERVER
        MENUITEM SEPARATOR
        MENUITEM "Quit",                        IDC_TASKBAR_QUIT
    END
END


/////////////////////////////////////////////////////////////////////////////
//
// Dialog
//

IDD_TASKBAR_STATUS DIALOG  0, 0, 145, 60
STYLE DS_SETFONT | DS_MODALFRAME | WS_POPUP
FONT 8, "MS Sans Serif"
BEGIN
    EDITTEXT        IDC_TASKBAR_STATUS_STATUS,3,3,139,12,ES_AUTOHSCROLL | ES_READONLY | NOT WS_BORDER
    LISTBOX         IDC_TASKBAR_STATUS_CLIENTS,3,17,139,40,NOT LBS_NOTIFY | LBS_SORT | LBS_NOINTEGRALHEIGHT | LBS_NOSEL | WS_VSCROLL | WS_TABSTOP
END


/////////////////////////////////////////////////////////////////////////////
//
// String Table
//

STRINGTABLE 
BEGIN
    IDS_FAILED              "Deskflow is about to quit with errors or warnings.  Please check the log then click OK."
    IDS_INIT_FAILED         "Deskflow failed to initialize: %{1}"
    IDS_UNCAUGHT_EXCEPTION  "Uncaught exception: %{1}"
END

#endif    // English (U.S.) resources
/////////////////////////////////////////////////////////////////////////////



#ifndef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 3 resource.
//


/////////////////////////////////////////////////////////////////////////////
#endif    // not APSTUDIO_INVOKED

