/*
 * Deskflow -- mouse and keyboard sharing utility
 * Copyright (C) 2024 Symless Ltd.
 *
 * This package is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * found in the file LICENSE that should have accompanied this file.
 *
 * This package is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include <QByteArray>
#include <QDataStream>
#include <QIODevice>
#include <QtGlobal>

namespace deskflow::gui {

inline int bytesToInt(const char *buffer, size_t size) {
  QByteArray byteArray(buffer, static_cast<int>(size));
  QDataStream stream(byteArray);
  int result;
  stream >> result;
  return result;
}

inline QByteArray intToBytes(int value) {
  QByteArray bytes;
  QDataStream stream(&bytes, QIODevice::WriteOnly);
  stream << value;
  return bytes;
}

} // namespace deskflow::gui
